import numpy as np

import noise3d
from .opr import dt, dv, dh, idt, idv, idh, NAMES
from .opr import n_s, n_t, n_v, n_h, n_tv, n_th, n_vh, n_tvh
from .opr import n_dt, n_dv, n_dh, n_dtdv, n_dtdh, n_dvdh


def _compute_spectrum_correction_matrix(T, V, H):
    """
    Compute spectrum mixin correction matrix
    """
    mat = 1/(T*V*H) * np.array(
        [
            [V*H*(T-1), 0, 0, H*(T-1), V*(T-1), 0, T-1],
            [0, T*H*(V-1), 0, H*(V-1), 0, T*(V-1), V-1],
            [0, 0, T*V*(H-1), 0, V*(H-1), T*(H-1), H-1],
            [0, 0, 0, H*(T-1)*(V-1), 0, 0, (T-1)*(V-1)],
            [0, 0, 0, 0, V*(T-1)*(H-1), 0, (T-1)*(H-1)],
            [0, 0, 0, 0, 0, T*(V-1)*(H-1), (V-1)*(H-1)],
            [0, 0, 0, 0, 0, 0, (T-1)*(V-1)*(H-1)],
        ]
    )
    return mat


def compute_meas_psd(seq):
    """
    Compute measured psd.
    Used for matrix approach.
    """
    T, V, H = seq.shape
    seq = seq - np.mean(seq)
    data_fft = np.fft.fftn(seq, axes=(0,1,2))
    data_fft_mod2 = np.real(data_fft * np.conjugate(data_fft))

    psd_tm = np.zeros(data_fft_mod2.shape)#, dtype=np.complex)
    psd_vm = np.zeros(data_fft_mod2.shape)#, dtype=np.complex)
    psd_hm = np.zeros(data_fft_mod2.shape)#, dtype=np.complex)
    psd_tvm = np.zeros(data_fft_mod2.shape)#, dtype=np.complex)
    psd_thm = np.zeros(data_fft_mod2.shape)#, dtype=np.complex)
    psd_vhm = np.zeros(data_fft_mod2.shape)#, dtype=np.complex)
    psd_tvhm = np.zeros(data_fft_mod2.shape)#, dtype=np.complex)

    psd_tm[:, 0, 0] = data_fft_mod2[:, 0, 0]
    psd_vm[0, :, 0] = data_fft_mod2[0, :, 0]
    psd_hm[0, 0, :] = data_fft_mod2[0, 0, :]
    psd_tvm[:, :, 0] = data_fft_mod2[:, :, 0] 
    psd_tvm = psd_tvm - psd_tm - psd_vm
    psd_thm[:, 0, :] = data_fft_mod2[:, 0, :]
    psd_thm = psd_thm - psd_tm - psd_hm
    psd_vhm[0, :, :] = data_fft_mod2[0, :, :]
    psd_vhm = psd_vhm - psd_vm - psd_hm
    psd_tvhm = data_fft_mod2 - (psd_tm + psd_vm + psd_hm + psd_tvm + psd_thm + psd_vhm)

    return psd_tm, psd_vm, psd_hm, psd_tvm, psd_thm, psd_vhm, psd_tvhm


def compute_psd(seq, names=False):
    """
    Compute noises psd.
    Relies on compute_meas_psd.
    """
    T, V, H = seq.shape
    correction_mat = _compute_spectrum_correction_matrix(T, V, H)
    inv_mat = np.linalg.inv(correction_mat)
    psd_m = np.asarray(compute_meas_psd(seq))
    res = tuple(np.einsum("ij,jklm->iklm", inv_mat, psd_m))
    return res + (NAMES, ) if names else res


def compute_var(seq, names=False):
    """
    Compute noise variances using spectrums.
    Using the matrix mixin approach.
    Relis on compute_psd.
    """
    T, V, H = seq.shape
    vec_psd = compute_psd(seq)
    vec_var = np.sum(vec_psd, axis=(1,2,3))/(T*V*H)**2
    t, v, h, tv, th, vh, tvh = vec_var
    res = t, v, h, tv, th, vh, tvh, np.sum(vec_var) # **2 : 1 DFT conventon, 1 to go into power units
    return res + (NAMES + ("tot",), ) if names else res


def var_psd_astrid(seq, numeric_invert=False, names=False):
    """Stand-alone method to compute bias corrected variances.
    
    Gives same values as corrected matrix method.
    
    numeric_invert allows to numericaly invert the mixing matrix. Faster if false.
    
    """
    #seq = seq - np.mean(seq)
    # Linear system matrix between sums (Es) and variances
    T, V, H = seq.shape
    
    # FT and norm of sequence
    psd_seq = np.fft.fftn(seq, axes=(0,1,2))
    norm_psd = np.multiply(psd_seq, np.conjugate(psd_seq))
    if not np.all(np.isreal(norm_psd)):
        raise ValueError("norm should be real")
    norm_psd = norm_psd.astype(np.float128, copy=False)
    
    # Generate indexes of cube sequence
    vt = np.linspace(0, T-1, T)
    vv = np.linspace(0, V-1, V)
    vh = np.linspace(0, H-1, H)
    mt, mv, mh  = np.meshgrid(vt, vv, vh, indexing="ij")
    
    # Sum over specific 3D region
    E3 = np.sum(norm_psd[np.logical_and(np.logical_and(mt!=0, mv!=0), mh!=0)])
    E4 = np.sum(norm_psd[np.logical_and(np.logical_and(mt==0, mv!=0), mh!=0)])
    E5 = np.sum(norm_psd[np.logical_and(np.logical_and(mt!=0, mv==0), mh!=0)])
    E6 = np.sum(norm_psd[np.logical_and(np.logical_and(mt!=0, mv!=0), mh==0)])
    E7 = np.sum(norm_psd[np.logical_and(np.logical_and(mt!=0, mv==0), mh==0)])
    E8 = np.sum(norm_psd[np.logical_and(np.logical_and(mt==0, mv!=0), mh==0)])
    E9 = np.sum(norm_psd[np.logical_and(np.logical_and(mt==0, mv==0), mh!=0)])

    # Solve linear system for variances
    Ev = np.array([E3, E4, E5, E6, E7, E8, E9])
    
    if numeric_invert:
        mat = (T*V*H) * np.array(
            [
            [0,     0,      0,      0,      0,      0,      (T-1)*(V-1)*(H-1)],
            [0,     0,      0,      0,      0,      T*(V-1)*(H-1),(V-1)*(H-1)],
            [0,     0,      0,      0,      V*(T-1)*(H-1),      0,      (T-1)*(H-1)],
            [0,     0,      0,      H*(T-1)*(V-1),      0,      0,      (T-1)*(V-1)],
            [(T-1)*(V)*(H),     0,      0,      (T-1)*H,      (T-1)*V,      0,      (T-1)],
            [0,     (V-1)*(T)*(H),      0,      (V-1)*H,      0,      (V-1)*T,      (V-1)],
            [0,     0,      (H-1)*(T)*(V),      0,      (H-1)*V,      (H-1)*T,      (H-1)]
            ]
        )
        var_psd = np.matmul(np.linalg.inv(mat), Ev)
    else:
        inv_mat = np.array(
            [
                [1/(H**2*T*V**2*(H - 1)*(T - 1)*(V - 1)), 0, -1/(H**2*T*V**2*(H - 1)*(T - 1)), -1/(H**2*T*V**2*(T - 1)*(V - 1)), 1/(H**2*T*V**2*(T - 1)), 0, 0],
                [1/(H**2*T**2*V*(H - 1)*(T - 1)*(V - 1)), -1/(H**2*T**2*V*(H - 1)*(V - 1)), 0, -1/(H**2*T**2*V*(T - 1)*(V - 1)), 0, 1/(H**2*T**2*V*(V - 1)), 0],
                [1/(H*T**2*V**2*(H - 1)*(T - 1)*(V - 1)), -1/(H*T**2*V**2*(H - 1)*(V - 1)), -1/(H*T**2*V**2*(H - 1)*(T - 1)), 0, 0, 0, 1/(H*T**2*V**2*(H - 1))],
                [  -1/(H**2*T*V*(H - 1)*(T - 1)*(V - 1)), 0, 0, 1/(H**2*T*V*(T - 1)*(V - 1)), 0, 0, 0],
                [  -1/(H*T*V**2*(H - 1)*(T - 1)*(V - 1)), 0, 1/(H*T*V**2*(H - 1)*(T - 1)), 0, 0, 0, 0],
                [  -1/(H*T**2*V*(H - 1)*(T - 1)*(V - 1)), 1/(H*T**2*V*(H - 1)*(V - 1)), 0, 0, 0, 0,0],
                [      1/(H*T*V*(H - 1)*(T - 1)*(V - 1)), 0, 0, 0, 0, 0, 0]
            ]
            )
        var_psd = np.matmul(inv_mat, Ev)

    t, v, h, tv, th, vh, tvh = var_psd
    res = t, v, h, tv, th, vh, tvh, np.sum(var_psd)
    return res + (NAMES + ("tot",), ) if names else res



class SpectrumDemo():
    
    def __init__(self, T=100, V=100, H=100, sigma=1, zero_mean=False):
        self.T = T
        self.V = V
        self.H = H
        self.sigma = sigma
        seq = noise3d.genseq.genseq_tvh(self.T,
                                        self.V,
                                        self.H,
                                        0,
                                        self.sigma)
        if zero_mean:
            seq = seq - np.mean(seq)
        self.seq = seq
        self._compute_psds()
    def _compute_psds(self):
        self.computed_psds = noise3d.spectrum.compute_psd(self.seq)
        self.psd = np.sum(self.computed_psds, axis=0)
        
        # Compute 3d fft
        self.dft = np.fft.fftn(self.seq, axes=(0, 1, 2))
        # take module2
        self.mod2 = np.real(np.conjugate(self.dft)*self.dft)
        
    def DFT_property_origin_value_vs_sum(self):
        """Compare the sum of the sequence with the square root of
        squared module of Fourier Transform of seq: 
            - sum(raw_seq)
            - np.sqrt(mod2[0, 0, 0])
        with mod2 = real(conj(dft(raw_seq))*dft(raw_seq))
        
        According to DFT properties, the origin value of the mod2 is the squared sum of the input signal : 
        
            mod2(0) = |\sum_{n}x_n e^{0}|^2 = |\sum_{n}x_n|^2
        
        """
        print("Sum of raw_seq : np.sum(seq):")
        print(np.sum(self.seq))
        print("Square root of mod2 of seq dft :")
        print(np.sqrt(self.mod2[0, 0, 0]))
        
    def DFT_property_mean_mod2_vs_sum_of_squared(self):
        """Compare the mean of the module and the sum of the squared :
            - np.sum(seq**2)
            - np.mean(mod2)
        
        According to DFT properties, the mean of the module2 is equal to the sum of the squared input signal.
        """
        print("Sum of squared raw_seq : np.sum(raw_seq**2)")
        print(np.sum(self.seq**2))
        print("Mean of mod2 : ")
        print(np.mean(self.mod2))
        
    def DFT_property_relation_to_variance(self):
        """Compare : 
            - Mean of mod2
            - Sum of squared input
            - N * variance of input
            
        In addition to mean(Mod2) == Sum(seq**2) in all case,
        if the input is 0-mean, it is also equal to N * variance(seq).
        
        Other way to put it : if the signal is 0-mean, 
        the variance is directly the sum of squared signal, 
        which is always the mean of mod2.
        """
        if not np.mean(self.seq)==0:
            raw_seq_0mean = self.seq - np.mean(self.seq)
        else:
            raw_seq_0mean = self.seq
        
        dft_0mean = np.fft.fftn(raw_seq_0mean, axes=(0, 1, 2))
        mod2_0mean = np.real(np.conjugate(dft_0mean)*dft_0mean)
        print("Remove mean value of raw_seq...")
        print("Sum of squared raw_seq : np.sum(raw_seq**2)")
        print(np.sum(raw_seq_0mean**2))
        print("Mean of mod2 : ")
        print(np.mean(mod2_0mean))
        print("N times Variance of seq")
        print(raw_seq_0mean.size * np.var(raw_seq_0mean))
        
    def DFT_property_Parseval(self):
        """
        The mean of the module is the sum of squared signal.
        
        If the signal is 0-mean, the variance can be used, as : 
            var(seq) = 1/N sum x_n^2
        """
        print("Mean of mod2")
        print(np.mean(self.mod2))
        print("Sum of squared signal")
        print(np.sum(self.seq**2))
        print("N times Variance of 0-mean seq")
        print(self.seq.size * np.var(self.seq-np.mean(self.seq)))
        
    def sums_of_psds(self):
        """
        Compare the sum of the DFT with the sum of
        the overall psd computed.
        
        """
        sum1 = np.sum(self.mod2)
        sum2 = np.sum(self.psd)
        print("Sum of mod2")
        print(sum1)
        print("Sum of psd")
        print(sum2)
        print(f"Error : {(sum1/sum2-1)}")
        
    def vars_of_pdsd(self):
        """Compare the variances of the input seq dft mod2 and 
        the recomputed psd. 
        If the sums are equals, then the 
        variance should be as well since the input seq must
        be 0-mean."""
        print(np.var(self.mod2))
        print(np.var(self.psd))
        
    def origin_value_close_to_zero(self):
        if not np.isclose(np.mean(self.seq), 0):
            raise ValueError("The input seq must be 0 mean for this comparison to work. This is because the way the 7 psds are computed first remove the mean.")
        print(self.psd[0, 0, 0])
        print(self.mod2[0, 0,0])

    def histograms(self, bins=100):
        """Overlay the 2 histograms"""
        import matplotlib.pyplot as plt
        plt.hist(self.psd.flatten(), bins=bins,
                 alpha=0.5, edgecolor="k",
                label="recomputed psd")
        plt.hist(self.mod2.flatten(), bins=bins,
                 alpha=0.5, edgecolor="k",
                 label="mod2")
        plt.legend()
        
    def diff_hist(self, bins=100):
        import matplotlib.pyplot as plt
        plt.hist(self.psd.flatten()-self.mod2.flatten(), bins=bins)


import numpy as np

from . import noise
from . import genseq

def var_t_err_tvh(T, V, H, var):
    """Compute the expected variance of t noise for a tvh sequence."""
    return var/(H*V)*(1 - 1/T)

def var_v_err_tvh(T, V, H, var):
    """Compute the expected variance of v noise for a tvh sequence."""
    return var/(H*T)*(1 - 1/V)


def var_h_err_tvh(T, V, H, var):
    """Compute the expected variance of h noise for a tvh sequence."""
    return var/(T*V)*(1 - 1/H)


def var_tv_err_tvh(T, V, H, var):
    """Compute the expected variance of tv noise for a tvh sequence."""
    return var/H*(1 - 1/T - 1/V + 1/(T*V))


def var_th_err_tvh(T, V, H, var):
    """Compute the expected variance of th noise for a tvh sequence."""
    return var/V*(1 - 1/T - 1/H + 1/(T*H))


def var_vh_err_tvh(T, V, H, var):
    """Compute the expected variance of vh noise for a tvh sequence."""
    return var/T*(1 - 1/V - 1/H + 1/(V*H))


def var_tvh_err_tvh(T, V, H, var):
    """Compute the expected variance of tvh noise for a tvh sequence."""
    return var*(1 - 1/T - 1/V - 1/H + 1/(T*V) + 1/(T*H) + 1/(V*H) - 1/(T*V*H))


class BiasedEstimatorTVH(object):
    """This class's purpose is to demonstrate the biasness of 
    the standard estimators Works only on tvh sequence.
    """
    def __init__(self, V=100, H=100, nb_matrix=100, sigma_tvh=1):
        
        self.t_min = 10
        self.t_max = 300
        self.delta_t = 5
        self.V = V
        self.H = H
        
        # Nb of iteration to average to improve estimation ("MonteCarlo")
        self.nb_matrix = nb_matrix
        
        # We choose a tvh sequence with std of 1
        self.sigma_tvh = sigma_tvh
    
    def setup(self):
        
        self.ech_t = np.arange(self.t_min, self.t_max, self.delta_t)
        
        # Setting up stack results
        self.res_var_tot = np.zeros((self.nb_matrix, len(self.ech_t)), dtype=np.float64)
        self.res_var_t = np.zeros((self.nb_matrix, len(self.ech_t)), dtype=np.float64)
        self.res_var_v = np.zeros((self.nb_matrix, len(self.ech_t)), dtype=np.float64)
        self.res_var_h = np.zeros((self.nb_matrix, len(self.ech_t)), dtype=np.float64)
        self.res_var_tv = np.zeros((self.nb_matrix, len(self.ech_t)), dtype=np.float64)
        self.res_var_th = np.zeros((self.nb_matrix, len(self.ech_t)), dtype=np.float64)
        self.res_var_vh = np.zeros((self.nb_matrix, len(self.ech_t)), dtype=np.float64)
        self.res_var_tvh = np.zeros((self.nb_matrix, len(self.ech_t)), dtype=np.float64)

    def compute_results(self):
        
        # "Monte Carlo" loop
        for k in range(1, self.nb_matrix+1):
            i = k-1
            # generate samples
            data_total = genseq.genseq_tvh(self.t_max, self.V, self.H, 0, self.sigma_tvh)
            j = 0
            # For all T-length
            for t in self.ech_t:
                # sub-sampling data on T axis
                data = data_total[0:t]
                # Compute the variances with given samples
                vt, vv, vh, vtv, vth, vvh, vtvh, vtot  = noise.get_all_3d_noise_var_fast(data)
                # Store results
                self.res_var_t[i, j] = vt
                self.res_var_v[i, j] = vv
                self.res_var_h[i, j] = vh
                self.res_var_tv[i, j] = vtv
                self.res_var_th[i, j] = vth
                self.res_var_vh[i, j] = vvh
                self.res_var_tvh[i, j] = vtvh
                self.res_var_tot[i, j] = vtot
                j = j+1

        # at this point, all data is stored in the res_var_ arrays.
        # Averaging results over all Monte Carlos
        self.mean_var_tot = np.mean(self.res_var_tot, axis=0, dtype=np.float64)
        self.mean_var_t = np.mean(self.res_var_t, axis=0, dtype=np.float64)
        self.mean_var_v = np.mean(self.res_var_v, axis=0, dtype=np.float64)
        self.mean_var_h = np.mean(self.res_var_h, axis=0, dtype=np.float64)
        self.mean_var_tv = np.mean(self.res_var_tv, axis=0, dtype=np.float64)
        self.mean_var_th = np.mean(self.res_var_th, axis=0, dtype=np.float64)
        self.mean_var_vh = np.mean(self.res_var_vh, axis=0, dtype=np.float64)
        self.mean_var_tvh = np.mean(self.res_var_tvh, axis=0, dtype=np.float64)
        
    def display_results(self):
        import matplotlib.pyplot as plt
        figure, axes = plt.subplots(2, 4, figsize=(14,6))
        # Computed variances labels
        var_labels = ('var_tot', 
                      'var_t', 'var_v', 'var_h',
                      'var_tv', 'var_th', 'var_vh',
                      "var_tvh")
        # Error labels
        error_labels = ("", 
                        'var_t_err_tvh', 'var_v_err_tvh', "var_h_err_tvh",
                        'var_tv_err_tvh', 'var_th_err_tvh', 'var_vh_err_tvh',
                        'var_tvh_err_tvh')
        # Computed variances values
        res = (self.mean_var_tot,
               self.mean_var_t, self.mean_var_v, self.mean_var_h,
               self.mean_var_tv, self.mean_var_th, self.mean_var_vh,
               self.mean_var_tvh)
        # Error functions
        error_funcs = (None,
                       var_t_err_tvh, var_v_err_tvh, var_h_err_tvh, 
                       var_tv_err_tvh, var_th_err_tvh, var_vh_err_tvh,
                       var_tvh_err_tvh)
        
        
        # Plotting 
        for ax, var_label, error_label, res_var, err_func in zip(axes.flatten(),
                                                                   var_labels,
                                                                   error_labels,
                                                                   res, 
                                                                   error_funcs):
            ax.plot(self.ech_t, res_var, label=var_label)
            if var_label != "var_tot":
                # no error function to plot on total variance
                ax.plot(self.ech_t, err_func(self.ech_t, self.V, self.H, self.sigma_tvh**2), "o", label=error_label)
            ax.legend()

        plt.tight_layout()



import numpy as np
import unittest

from noise3d import genseq

class TestGenseq(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        cls.T = 100
        cls.V = 100
        cls.H = 100
            
    def test_10_1d_seqs(cls):
        # checking dims and shapes
        cls.assertEqual(genseq.genseq_t(cls.T, cls.V, cls.H, 0, 1).shape, (cls.T, cls.V, cls.H))
        cls.assertEqual(genseq.genseq_v(cls.T, cls.V, cls.H, 0, 1).shape, (cls.T, cls.V, cls.H))
        cls.assertEqual(genseq.genseq_h(cls.T, cls.V, cls.H, 0, 1).shape, (cls.T, cls.V, cls.H))
        
    def test_20_2d_seqs(cls):
        # checking dims and shapes
        cls.assertEqual(genseq.genseq_tv(cls.T, cls.V, cls.H, 0, 1).shape, (cls.T, cls.V, cls.H))
        cls.assertEqual(genseq.genseq_th(cls.T, cls.V, cls.H, 0, 1).shape, (cls.T, cls.V, cls.H))
        cls.assertEqual(genseq.genseq_vh(cls.T, cls.V, cls.H, 0, 1).shape, (cls.T, cls.V, cls.H))

    def test_30_3d_seq(cls):
        # checking dims and shapes
        cls.assertEqual(genseq.genseq_tvh(cls.T, cls.V, cls.H, 0, 1).shape, (cls.T, cls.V, cls.H))
        
    def test_40_gen_seqs(cls):
        mus = tuple([0 for i in range(7)])
        sigmas = tuple([1 for i in range(7)])
        
        seqs = genseq.genseq_all_seq(cls.T, cls.V, cls.H, sigmas, mus=mus)
        
        # checking dims and shapes
        for seq in seqs:
            cls.assertEqual(seq.shape, (cls.T, cls.V, cls.H))

        # checking the position of the total noise sequence
        cls.assertTrue(np.all(seqs[-1] == sum(seqs[:-1])))
        

if __name__ == "__main__":
    unittest.main()
        
        


"""
Generate noise sequences according to the 3D noise model.

This module allows to create 3D sequences of T frames, 
V lines and H columns, with any combination of the 7 noise types, namely : 
 - temporal noise                     : t   (flicker)
 - vertical noise                     : v   (fixed lining)
 - horizontal noise                   : h   (fixed columning)
 - temporal-horizontal noise          : tv  (temporaly varying lining)
 - temporal-vertical noise            : th  (temporaly varying columning)
 - vertical-horizontal noise          : vh  (fixed pattern noise)
 - temporal-vertical-horizontal noise : tvh (3d-uncorrelated noise)

Each kind of noise can be created with mean and standard 
deviation values as, for eg :

    # 0-mean, 1 standard deviation
    flicker_seq = genseq_t(T, V, H, 0, 1)
 
All of the 7 noise sequences can be created in one call, with tuples
of standard deviation and means (sigmas and mus) : 

    n_seqs = genseq_all_seq(T, V, H, sigmas, mus)
    arr_t, arr_v, arr_h, arr_tv, arr_th, arr_vh, arr_tvh, arr_tot = n_seqs

By default, mus are 0s.

Finaly, you can get the 3D sequence directly with : 

    noises_seq = noise3d.genseq.genseq_3dnoise_seq(T, V, H, sigmas)


TODO/questions : 
 - allow order types of noise ? for now only random normal
 - should always keepdims ? allow as argument ? set as package config variable ?
 - allow order changing dimensions ? (T, V, H by default)
 - define all default means to 0 and all sigmas to 1 ?
"""


import numpy as np


MUS = (0, 0, 0, 0, 0, 0, 0)
NAMES = ('t', 'v', 'h', 'tv', 'th', 'vh', 'tvh')


def genseq_t(T, V, H, mu, sigma):
    """Generate a 3D T-V-H sequence of gaussian t-noise (flicker) with mean mu and std sigma."""
    val = np.random.normal(mu, sigma, T)
    arr = np.repeat(np.repeat(val[:, np.newaxis], V, axis=1)[:, :, np.newaxis], H, axis=2)
    return arr


def genseq_v(T, V, H, mu, sigma):
    """Generate a 3D T-V-H sequence of gaussian v-noise with mean mu and std sigma."""
    val = np.random.normal(mu, sigma, V)
    arr = np.repeat(np.repeat(val[np.newaxis, :], T, axis=0)[:, :, np.newaxis], H, axis=2)
    return arr


def genseq_h(T, V, H, mu, sigma):
    """Generate a 3D T-V-H sequence of gaussian h-noise with mean mu and std sigma."""
    val = np.random.normal(mu, sigma, H)
    arr = np.repeat(np.repeat(val[np.newaxis, :], T, axis=0)[:, np.newaxis, :], V, axis=1)
    return arr


def genseq_tv(T, V, H, mu, sigma):
    """Generate a 3D T-V-H sequence of gaussian tv-noise with mean mu and std sigma."""
    val = np.random.normal(mu, sigma, T*V)
    val = np.reshape(val, (T, V))
    arr = np.repeat(val[:, :, np.newaxis], H, axis=2)
    return arr


def genseq_th(T, V, H, mu, sigma):
    """Generate a 3D T-V-H sequence of gaussian th-noise with mean mu and std sigma."""
    val = np.random.normal(mu, sigma, T*H)
    val = np.reshape(val, (T, H))
    arr = np.repeat(val[:, np.newaxis, :], V, axis=1)
    return arr


def genseq_vh(T, V, H, mu, sigma):
    """Generate a 3D T-V-H sequence of gaussian vh-noise (fixed pattern noise) with mean mu and std sigma."""
    val = np.random.normal(mu, sigma, V*H)
    val = np.reshape(val, (V, H))
    arr = np.repeat(val[np.newaxis, :, :], T, axis=0)
    return arr


def genseq_tvh(T, V, H, mu, sigma):
    """Generate a 3D T-V-H sequence of gaussian tvh-noise ("photon" noise) with mean mu and std sigma."""
    val = np.random.normal(mu, sigma, T*V*H)
    arr = np.reshape(val, (T,V,H))
    return arr


def genseq_all_seq(T, V, H, sigmas, mus=MUS, names=False):
    """Generate all 7 3D T-V-H sequences of noises, and return them as well as the total sequence :
     - t sequence
     - v sequence
     - h sequence
     - tv sequence
     - th sequence
     - vh sequence
     - tvh sequence
     - sum of all of these
     The corresponding std and means must be passed in a tuple. Default means are 0 for all noise.
     """
    sigma_t, sigma_v, sigma_h, sigma_tv, sigma_th, sigma_vh, sigma_tvh = sigmas
    mu_t, mu_v, mu_h, mu_tv, mu_th, mu_vh, mu_tvh = mus
    arr_t   = genseq_t(T, V, H, mu_t, sigma_t)
    arr_v   = genseq_v(T, V, H, mu_v, sigma_v)
    arr_h   = genseq_h(T, V, H, mu_h, sigma_h)
    arr_tv  = genseq_tv(T, V, H, mu_tv, sigma_tv)
    arr_th  = genseq_th(T, V, H, mu_th, sigma_th)
    arr_vh  = genseq_vh(T, V, H, mu_vh, sigma_vh)
    arr_tvh = genseq_tvh(T, V, H, mu_tvh, sigma_tvh)
    arr_tot = np.sum([arr_t, arr_v, arr_h, arr_tv, arr_th, arr_vh, arr_tvh], axis=0)
    seqs = arr_t, arr_v, arr_h, arr_tv, arr_th, arr_vh, arr_tvh, arr_tot
    return  seqs + (NAMES+('tot',), ) if names else seqs


def genseq_3dnoise_seq(T, V, H, sigmas, mus=MUS):
    """Generate a 3D T-V-H sequence with all noises with standard-deviation sigmas and means mus.
     The corresponding std and means must be passed in a tuple. Default means are 0 for all noise.
     """
    return genseq_all_seq(T, V, H, sigmas, mus)[-1]



import numpy as np

import matplotlib.pyplot as plt

import noise3d
from . import noise
from . import opr
from . import stats
from . import spectrum
#from . import spectrum as ns

#import scipy.stats as st

# ORDER is used to re-order the sequences to match TITLES (for display)
ORDER = [-1, 1, 2, 5, 0, 3, 4, 6]
TITLES = ["tot", "v", "h", "vh", "t", "tv", "th", "tvh"]


def gauss(x, mu, sigma):
    """Helper gaussian function"""
    return np.exp(-(x-mu)**2/(2*sigma**2))/np.sqrt(2*np.pi*sigma**2)


def histo_seq(seq, ax=None, fit=True, stats_on=True, stats_print=False,
              print_CI=False, nbin=10,
              density=True):
    """
    Plot histrogram of seq with additional infos.
    """
    if ax is None:
        fig, ax = plt.subplots(1, 1)
    # Compute histogram
    vals = seq.flatten()
    ech = np.linspace(np.min(seq), np.max(seq), nbin)
    val, bins, patchs = ax.hist(vals, bins=ech, density=density)
    # Compute basic stats
    mean = np.mean(vals)
    std = np.std(vals)
    
    # Add a gaussian fit
    if fit:
        ech_gauss = gauss(ech, mean, std)
        ax.plot(ech, ech_gauss)

    # If stats are needed
    if stats_print or stats_on:
        results = stats.basic_stats(seq)
        
    # Print stats in standard output
    if stats_print:
        print("Count : {}".format(results["count"]))
        print("Min/Max : {}/{}".format(results["min"], results["max"]))
        print("Mean : {}".format(results["mean"]))
        print("Median : {}".format(results["median"]))
        print("Var : {}".format(results["var"]))
        print("Std : {}".format(results["std"]))

    # Add stat on plot
    if stats_on:
        ax.errorbar(mean,
                    max(val)/2,
                    yerr=None,
                    xerr=std,
                    solid_capstyle='projecting',
                    capsize=5)
        text = "Count : {}\nMin/Max : {:.3f}/{:.3f}\nMean : {:.3f}\nMedian : {:.3f}\nVar : {:.3f}\nStd : {:.3f}".format(
            results["count"], 
            results["min"], 
            results["max"],
            results["mean"],
            results["median"],
            results["var"],
            results["std"])
        
        ax.annotate(text, xy=(0.99, 0.99),
                    xycoords='axes fraction', 
                    xytext=(0.99, 0.99), 
                    textcoords='axes fraction',
                    horizontalalignment='right',
                    verticalalignment='top')

    return ax


def display_8seq(seqs, extract=False, fit=True, stats_on=True, 
                 stats_print=False, print_CI=False,
                 nbin=50, density=True, figsize=(8.0, 6.0),
                 samex=False):
    """Display all 8 histograms.
    Input seqs can be :
         - 1 raw 3d sequence, with extract=True
         - 8 sequences in a list/tuple (t, v, h, tv, th, vh, tvh, tot)
    """
    # analyze sequences 
    if extract:
        seqs = noise3d.opr.get_all_3d_noise_seq_fast(seqs, names=True)
        names = seqs[-1]
        seqs = seqs[:-1]
    # define plots
    fig, axes = plt.subplots(2, 4, figsize=figsize)
    axes = axes.flatten()
    # Compute histgram bounds if samex
    if samex:
        xmin = np.min(seqs[-1])
        xmax = np.max(seqs[-1])
    # reorder sequences
    seqs = [seqs[i] for i in ORDER]
    # for all sequences
    for seq, ax, title in zip(seqs, axes, TITLES):
        histo_seq(seq, ax=ax, fit=fit, stats_on=stats_on,
                  stats_print=stats_print,
                  print_CI=print_CI, nbin=nbin, 
                  density=density)
        # setting title
        ax.set_title(title)
        # seetting x axis limits
        if samex:
            ax.set_xlim(xmin, xmax)
    fig = plt.gcf()
    fig.tight_layout()
    return fig



def noise_resume(seq, pc=False, method="fast"):
    """
    Compute all noise infos in a string
    """
    T, V, H = seq.shape
    # compute noise
    noises = noise.get_noise_vars(seq, method=method)
    # express all noise in % of total variance
    if pc: 
        noises = tuple(np.array(noises)/noises[-1])

    # split noises tuple
    var_t, var_v, var_h, var_tv, var_th, var_vh, var_tvh, var_tot = noises
        
    sep = "-------------------------------------------------\n"
    
    # general infos
    high_lvl = "Mean : {:.3f} | Var : {:.3f} | Sum : {:.3f}\n".format(np.mean(seq), np.var(seq), np.sum(seq))
    shape = "T : {} | V : {} |  H : {} | Max error : {:.1%}\n".format(T, V, H, ((1-1/T-1/V-1/H)-1))
    method = "---------------- {:10} ---------------------\n".format(method)
    # line of noise
    res_1 = '$\sigma^2_tot$ : {:6.3f} | $\sigma^2_v$ :  {:6.3f} | $\sigma^2_h$ :  {:6.3f} | $\sigma^2_vh$ :  {:6.3f}\n'.format(var_tot, var_v, var_h, var_vh)
    # line 2 of noise
    res_2 = '$\sigma^2_t$ : {:8.3f} | $\sigma^2_tv$ : {:6.3f} | $\sigma^2_th$ : {:6.3f} | $\sigma^2_tvh$ : {:6.3f} \n'.format(var_t, var_tv, var_th, var_tvh)
    #concatenate
    string_resume = sep + high_lvl + shape + method + res_1 + res_2 + sep
    return string_resume


### Display spectrum
def disp_spectrum(seq, 
                  xt=0, xv=0, xh=0,
                  figsize=(4,2),
                  share_scale=True):
    from matplotlib import cm
    # Compute spectrums
    vec_psds = spectrum.compute_psd(seq)
    t, v, h, tv, th, vh, tvh = vec_psds
    
    # get bounds
    vmin_vec = np.min(vec_psds)
    vmax_vec = np.max(vec_psds)

    # Extract 1D spectrums
    psd_t_1D = t[:, xv, xh]
    psd_v_1D = v[xt, :, xh]
    psd_h_1D = h[xt, xv, :]
    # Extract 2D spectrums
    psd_tv_2D = tv[:, :, xh]
    psd_th_2D = th[:, xv, :]
    psd_vh_2D = vh[xt, :, :]
    # Extract 3D spectrum
    psd_tvh_3D = tvh

    # create plots
    fig, axes = plt.subplots(3, 3, figsize=figsize)
    axes = axes.ravel().tolist()
        
    # Plot 1D spectrum
    #axes[0].plot(psd_t_1D)
    axes[0].scatter(np.arange(len(psd_t_1D)), psd_t_1D, c=cm.viridis(psd_t_1D/np.max(psd_t_1D)), edgecolor="none")
    axes[0].set_xlabel("t")
    axes[0].set_title("t noise")
    #axes[1].plot(psd_v_1D)
    axes[1].scatter(np.arange(len(psd_v_1D)), psd_v_1D, c=cm.viridis(psd_v_1D/np.max(psd_v_1D)), edgecolor="none")
    axes[1].set_xlabel("v")
    axes[1].set_title("v noise")
    #axes[2].plot(psd_h_1D)
    axes[2].scatter(np.arange(len(psd_h_1D)), psd_h_1D, c=cm.viridis(psd_h_1D/np.max(psd_h_1D)), edgecolor="none")
    axes[2].set_xlabel("h")
    axes[2].set_title("h noise")
    if share_scale:
        axes[0].set_ylim((vmin_vec, vmax_vec))
        axes[1].set_ylim((vmin_vec, vmax_vec))
        axes[2].set_ylim((vmin_vec, vmax_vec))
    
    # Plot 2D spectrum
    if share_scale:
        axes[3].imshow(psd_tv_2D, vmin=vmin_vec, vmax=vmax_vec)
        axes[4].imshow(psd_th_2D, vmin=vmin_vec, vmax=vmax_vec)
        axes[5].imshow(psd_vh_2D, vmin=vmin_vec, vmax=vmax_vec)
    else:
        axes[3].imshow(psd_tv_2D)
        axes[4].imshow(psd_th_2D)
        axes[5].imshow(psd_vh_2D)
    axes[3].set_title("tv noise")
    axes[3].set_xlabel("t")
    axes[3].set_ylabel("v")
    axes[4].set_title("th noise")
    axes[4].set_xlabel("t")
    axes[4].set_ylabel("h")
    axes[5].set_title("vh noise")
    axes[5].set_xlabel("v")
    axes[5].set_ylabel("h")
    
    # Plot 3D spectrum
    if share_scale:
        axes[6].imshow(psd_tvh_3D[xt, :, :],
                       vmin=vmin_vec, vmax=vmax_vec)
        axes[7].imshow(psd_tvh_3D[:, xv, :],
                       vmin=vmin_vec, vmax=vmax_vec)
        im = axes[8].imshow(psd_tvh_3D[:, :, xh],
                            vmin=vmin_vec, vmax=vmax_vec)
    else:
        axes[6].imshow(psd_tvh_3D[xt, :, :])
        axes[7].imshow(psd_tvh_3D[:, xv, :])
        im = axes[8].imshow(psd_tvh_3D[:, :, xh])
    
    # Titles and labels
    axes[6].set_title(f"tvh noise - vh plane (t={xt})")
    axes[6].set_xlabel("v")
    axes[6].set_ylabel("h")
    axes[7].set_title(f"tvh noise - th plane (v={xv})")
    axes[7].set_xlabel("t")
    axes[7].set_ylabel("h")
    axes[8].set_title(f"tvh noise - tv plane (h={xh})")
    axes[8].set_xlabel("t")
    axes[8].set_ylabel("v")

    #fig.suptitle("Noise Power Spectrum Density")
    plt.tight_layout()
    fig.colorbar(im, ax=axes)

    return fig, axes




class SequenceViewer(object):
    """Allow to view 8 3d volumes.
    Use j and k keys to change slice on first axis (typicaly temporal).
    The 8 volumes must have same shapes.
    """
    
    def __init__(self, volumes, title_list=TITLES,
                 vmin="default", vmax="default",
                 cmap="gray", interpolation=None):
        self._remove_keymap_conflicts({'j', 'k'})
        
        fig, ax = plt.subplots(nrows=2, ncols=4, squeeze=True)
        self.fig = fig
        if vmin=="default" and vmax=="default":
            vmin = np.min(np.asarray(volumes))
            vmax = np.max(np.asarray(volumes))
        
        # reorder seqs to match classic display order 
        volumes = [volumes[i] for i in ORDER]
        
        for myax, volume, title in zip(ax.flatten(), volumes, title_list):
            myax.volume = volume
            myax.index = 0
            myax.seq_name = title
            im = myax.imshow(volume[myax.index], vmin=vmin, vmax=vmax, interpolation=interpolation, cmap=cmap)
            myax.set_title(title + "[{}]".format(myax.index))
            myax.axis("off")
        #fig.canvas.draw()
        plt.tight_layout()
        fig.canvas.mpl_connect('key_press_event', self._process_key)
        #fig.colorbar(im, ax=ax)
        fig.suptitle("3D noise sequences\n(use j and k)")
        
    
    def _remove_keymap_conflicts(self, new_keys_set):
        for prop in plt.rcParams:
            if prop.startswith('keymap.'):
                keys = plt.rcParams[prop]
                remove_list = set(keys) & new_keys_set
                for key in remove_list:
                    keys.remove(key)

    
    def _process_key(self, event):
        """Widget API"""
        fig = event.canvas.figure
        for ax in fig.axes:
            if event.key == 'j':
                self._change_slice(ax, "prev")
            elif event.key == 'k':
                self._change_slice(ax, "next")
            fig.canvas.draw()
            plt.tight_layout()
    
    # Doesnt work as it is
    #def change_all_slice(self, next_or_prev="next"):
    #    """Python API : viewer.change_all_slice()"""
    #    fig = self.fig
    #    for ax in fig.axes:
    #        self._change_slice(ax, next_or_prev=next_or_prev)
    #        fig.canvas.draw()
    #        plt.tight_layout()
    
    def _change_slice(self, ax, next_or_prev="next"):
        # update ax.index
        if next_or_prev == "next":
            ax.index = (ax.index + 1) % ax.volume.shape[0] # wrap around using %
        elif next_or_prev == "prev":
            ax.index = (ax.index - 1) % ax.volume.shape[0] # wrap around using %
        else:
            raise ValueError("next or prev not ", next_or_prev)
        # update data
        ax.images[0].set_array(ax.volume[ax.index])
        # update titles
        ax.set_title(ax.seq_name + "[{}]".format(ax.index))
        #plt.show()


def idea_to_display_3d_seq():
    """Starter for an interactive 3D display of a sequence"""
    #%matplotlib qt

    import numpy as np
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D
    
    N = 4
    volume = np.random.rand(N, N, N)
    filled = np.ones((N, N, N), dtype=np.bool)
    
    # repeating values 3 times for grayscale
    colors = np.repeat(volume[:, :, :, np.newaxis], 3, axis=3)
    
    fig = plt.figure()
    ax = fig.gca(projection='3d')
    
    ax.voxels(filled, facecolors=colors, edgecolors='k')
    plt.show()



"""
Basic statistics computation functions.
"""

import numpy as np
import matplotlib.pyplot as plt
import scipy.stats as st


def basic_stats(data, ddof=1):
    """
    Compute basic stats on data
    """
    # Compute basic stats
    d_count = data.size
    n = d_count
    d_min = np.min(data)
    d_max = np.max(data)
    d_median = np.median(data)
    d_mean = np.mean(data)
    d_var = np.var(data, ddof=ddof)
    d_std = np.std(data, ddof=ddof)

    results = {
        "min":d_min,
        "max":d_max,
        "median":d_median,
        "mean":d_mean,
        "var":d_var,
        "std":d_std,
        "count":d_count,
    }
    return results


def compute_CI(data, mu_X=None, sigma_X=None, law_X="normal", alpha=0.05, print_CI=False, ddof=1):
    """
    Compute Confidence Interval.
    """
    if not law_X=="normal":
        raise NotImplementedError("data must be gaussian.")
    
    d_count = data.size
    n = d_count
    d_mean = np.mean(data)
    # Mean
    ## If X follows normal lax
    if law_X=="normal":
        # In all case, Xbar follows a symetric lax
        alpha_Xbar_2 = alpha/2

        # If X variance is known
        if sigma_X is not None:
            # Compute std of Xbar law
            sigma_Xbar = sigma_X/np.sqrt(n)

            # Then Y=((Xbar-mu)/sigma_xbar) follows a N(0,1) law

            # Compute boundary for N(0,1) such that the area between boundaries is proba_Xbar%

            # Using the percent point function, such that ppf(proba) give the upper bound
            zscore_CI_p = st.norm.ppf(1 - alpha_Xbar_2)
            # Since the law is N(0,1), law is symetric so
            zscore_CI_m = - zscore_CI_p

            # Translate the bounds in terms of X such that mu is in Xbar +-sigma_Xbar, 
            # with probability proba_Xbar%
            ci_Xbar_p = zscore_CI_p * sigma_Xbar
            ci_Xbar_m = zscore_CI_m * sigma_Xbar

            # In the end, we know mu should be in XBar+-ci at proba_Xbar%
            if print_CI:
                print("\sigma_X known, using z-score to compute CI for Xmean")
                print("\mu is within {}+/-{} at {}%".format(d_mean, ci_Xbar_p, alpha))

        # If X variance is not known
        if sigma_X is None:
            # First, estimate sigma_X, and then Xbar follow a Student t law
            s_X = np.std(data, ddof=ddof)
            s_Xbar = s_X/np.sqrt(n)

            df_Xbar = n-1
            tscore_CI_p = st.t.ppf(1 - alpha_Xbar_2, df=df_Xbar)
            # Since we have a Studen law, it is symetric
            tscore_CI_m = - tscore_CI_p

            # Translate the bounds in terms of X such that mu is in Xbar +-sigma_Xbar at proba_Xbar% 
            ci_Xbar_p = tscore_CI_p * s_Xbar
            ci_Xbar_m = tscore_CI_m * s_Xbar
            if print_CI:
                print("\sigma_{X} unknow, estimating it and using Student's t distribution for Xmean")
                print("\mu is within {}+-{} at {}%".format(d_mean, ci_Xbar_p, alpha))


    # Variance
    ## Variance follows a Ki2 lax with df=n-1 dof
    if law_X=="normal":
        # If X mean is known
        if mu_X is not None:
            # s2 follows a ki2 with N degrees 
            s2 = np.mean(np.abs(data - mu_X)**2) * n/(n-ddof)

            # Compute reduced variable bounds 
            kiscore_CI_var_p = st.chi2.ppf(1 - alpha, df=n)
            kiscore_CI_var_m = st.chi2.ppf(alpha, df=n)

            # Translate in terms of variance
            ci_var_p = s2 *(n-ddof) / kiscore_CI_var_m
            ci_var_m = s2 *(n-ddof) / kiscore_CI_var_p
            if print_CI:
                print("\mu_X known, using chi2 score with df=n to compute CI for var")
                print("Var_x is wihtin {}+-{}/{} at {}%".format(s2, ci_var_m, ci_var_p, alpha))


        # If X mean is not knwon
        if mu_X is None:
            # s2 folllow a ki2 lax at N-1 degrees
            s2 = np.var(data, ddof=ddof)

            # Compute reduced variable bounds
            kiscore_CI_var_p = st.chi2.ppf(1 - alpha, df=n)
            kiscore_CI_var_m = st.chi2.ppf(alpha, df=n)

            # Translate in terms of variance
            ci_var_p = s2 *(n-ddof) / kiscore_CI_var_m
            ci_var_m = s2 *(n-ddof) / kiscore_CI_var_p
            if print_CI:
                print("\mu_X not known, using chi2 score with df=n-1 to compute CI for var")
                print("Var_x is wihtin {}+-{}/{} at {}%".format(s2, ci_var_m, ci_var_p, alpha))

    results = {
        "ci_Xbar_m":ci_Xbar_m,
        "ci_Xbar_p":ci_Xbar_p,
        "ci_var_m":ci_var_m,
        "ci_var_p":ci_var_p,
    }
    return results




import numpy as np
import unittest

import noise3d
from noise3d import noise

class TestGenseq(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        cls.T = 100
        cls.V = 100
        cls.H = 100
        vals = np.random.normal(0, 1, cls.T*cls.V*cls.H)
        cls.seq = np.reshape(vals, (cls.T,cls.V,cls.H))
        
    def test_10_basic_variances(cls):
        
        var_t = noise3d.noise.var_nt(cls.seq)
        var_v = noise3d.noise.var_nv(cls.seq)
        var_h = noise3d.noise.var_nh(cls.seq)
        var_tv = noise3d.noise.var_ntv(cls.seq)
        var_th = noise3d.noise.var_nth(cls.seq)
        var_vh = noise3d.noise.var_nvh(cls.seq)
        var_tvh = noise3d.noise.var_ntvh(cls.seq)
        singles = (var_t, var_v, var_h, 
                  var_tv, var_th, var_vh,
                  var_tvh)
        
        var_set = noise3d.noise.get_all_3d_noise_var_classic(cls.seq, names=True)
        names = var_set[-1]
        var_set = var_set[:-1]
        expected_names = ('t', 'v', 'h', 
                         'tv', 'th', 'vh',
                         'tvh')
        
        for single, var, name, exp_name in zip(singles, var_set, names, expected_names):
            cls.assertEqual(single, var)
            cls.assertEqual(name, exp_name)
        
    def test_20_fast_equivalent(cls):
        fast_vars = noise3d.noise.get_all_3d_noise_var_fast(cls.seq, names=True)
        classic_vars = noise3d.noise.get_all_3d_noise_var_classic(cls.seq, names=True)
        
        for fast, classic in zip(fast_vars, classic_vars):
            cls.assertAlmostEqual(fast, classic, places=4)
            # cls.assertEqual(fast, classic)
            # cls.assertAlmostEqual(fast, classic)
            
    def test_30_classic_matrix(cls):
        classic_approach = noise3d.noise.get_all_3d_noise_var_classic(cls.seq, names=True)
        matrix_approach = noise3d.noise.get_all_3d_classic_var_matrix(cls.seq, names=True)
        
        for classic, matrix in zip(classic_approach, matrix_approach):
            #cls.assertEqual(classic, matrix)
            cls.assertAlmostEqual(classic, matrix, places=4)
            
            
    def test_40_corrected_matrix(cls):
        corrected = noise3d.noise.get_all_3d_corrected_var_matrix(cls.seq)
        
        
if __name__ == "__main__":
    unittest.main()



from .genseq import genseq_3dnoise_seq, NAMES

from .opr import get_all_3d_noise_seq_classic

from .noise import get_all_3d_noise_var_classic, get_all_3d_noise_var_fast, get_all_3d_corrected_var_matrix

from . import disp


import numpy as np
import unittest

from noise3d import bias

class TestGenseq(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        cls.T = 100
        cls.V = 100
        cls.H = 100
        cls.var = 1
        
    def test_00_error_functions(cls):
        funcs = (bias.var_t_err_tvh, bias.var_v_err_tvh, bias.var_h_err_tvh,
                 bias.var_tv_err_tvh, bias.var_th_err_tvh, bias.var_vh_err_tvh, 
                 bias.var_tvh_err_tvh)
        for func in funcs:
            try:
                func(cls.T, cls.V, cls.H, cls.var)
            except:
                cls.fail()
        
    def test_10_demo(cls):
        demo = bias.BiasedEstimatorTVH(V=10, H=10, nb_matrix=20)
        try:
            demo.setup()
            demo.compute_results()
        except:
            cls.fail()
        

if __name__ == "__main__":
    unittest.main()



""""""

import numpy as np

from .genseq import NAMES

DTYPE = np.float64
MEAN_NAMES = ('dvdh','dtdh','dtdv','dh','dv','dt')


# Base operators
def dt(seq):
    """
    Means the 3D input sequence along the T axis.
    Remember the convention of this package : T on 0 axis, V on 1 axis, H on 2 axis.
    Returns a sequence with same dimension as input.
    """
    return np.mean(seq, axis=0, dtype=DTYPE, keepdims=True)


def dv(seq):
    """
    Means the 3D input sequence along the V axis.
    Remember the convention of this package : T on 0 axis, V on 1 axis, H on 2 axis.
    Returns a sequence with same dimension as input.
    """
    return np.mean(seq, axis=1, dtype=DTYPE, keepdims=True)


def dh(seq):
    """
    Means the 3D input sequence along the H axis.
    Remember the convention of this package : T on 0 axis, V on 1 axis, H on 2 axis.
    Returns a sequence with same dimension as input.
    """
    return np.mean(seq, axis=2, dtype=DTYPE, keepdims=True)


def idt(seq):
    """
    Filters the T noise of the 3D input sequence.
    Returns a sequence with same shape as input, 
    thanks to numpy's broadcast.
    """
    return seq - dt(seq)


def idv(seq):
    """
    Filters the V noise of the 3D input sequence.
    Returns a sequence with same shape as input, 
    thanks to numpy's broadcast.
    """
    return seq - dv(seq)


def idh(seq):
    """
    Filters the H noise of the 3D input sequence.
    Returns a sequence with same shape as input, 
    thanks to numpy's broadcast.
    """
    return seq - dh(seq)


# Noise sequence
## linear approach
def n_s(seq):
    """
    Extract the mean of the 3D input sequence,
    by applying all 3 base mean-operators.
    Equivalent to np.mean(seq).
    """
    return dt(dv(dh(seq)))


def n_t(seq):
    """
    Extract the temporal t noise sequence of the 3D 
    input sequence, with same shape.
    """
    return dv(dh(idt(seq)))


def n_v(seq):
    """
    Extract the vertical v noise sequence of the 3D 
    input sequence, with same shape.
    """
    return dt(dh(idv(seq)))


def n_h(seq):
    """
    Extract the horizontal h noise sequence of the 3D 
    input sequence, with same shape.
    """
    return dv(dt(idh(seq)))


def n_tv(seq):
    """
    Extract the temporel-vertical tv noise sequence of the 3D 
    input sequence, with same shape.
    """
    return dh(idv(idt(seq)))


def n_th(seq):
    """
    Extract the temporal-horizontal th noise sequence of the 3D 
    input sequence, with same shape.
    """
    return dv(idh(idt(seq)))


def n_vh(seq):
    """
    Extract the vertical-horizontal vh noise sequence of the 3D 
    input sequence, with same shape.
    """
    return dt(idv(idh(seq)))


def n_tvh(seq):
    """
    Extract the vertical temporel-vertical-horizontal noise 
    sequence of the 3D input sequence, with same shape.
    """
    return idh(idv(idt(seq)))


def get_all_3d_noise_seq_classic(seq, names=False):
    """
    Extract all 7 noise sequences of the 3D input sequence, with same shape.
    Returns the 7 sequences plus input seq in a tuple.
    """
    seqs = n_t(seq), n_v(seq), n_h(seq), n_tv(seq), n_th(seq), n_vh(seq), n_tvh(seq), seq
    return seqs + (NAMES+('tot', ), ) if names else seqs


#### Basic extraction of noise sequences
def get_all_3d_noise_seq_fast(seq, names=False):
    """
    Extract all 7 noise sequences of a 3D sequence.
    
    Faster version of get_all_3d_noise_seq(seq).
    
    This basic approach consists in filtering noise in some
    directions and extracting it from the base sequence.
    
    Returns all noise sequences in a tuple, as well as the 
    original sequence, in the following order : 
        (t, v, h, tv, th, vh, tvh, input)
        
    This function is called "fast" beacause it limits the 
    number of dt, dv, dh operators calls.
    In most cases, it should lead to quite accurate results.
    """
    # 7 base mean sequences
    seq_dt = dt(seq)
    seq_dv = dv(seq)
    seq_dh = dh(seq)
    seq_dtdv = dv(seq_dt)
    seq_dtdh = dh(seq_dt)
    seq_dvdh = dv(seq_dh)
    # overall mean
    seq_dtdvdh = dh(seq_dtdv)
    # 7 noise sequences
    seq_s = seq_dtdvdh
    seq_t = seq_dvdh - seq_dtdvdh
    seq_v = seq_dtdh - seq_dtdvdh
    seq_h = seq_dtdv - seq_dtdvdh
    seq_tv = seq_dh - seq_dtdh - seq_dvdh + seq_dtdvdh 
    seq_th = seq_dv - seq_dtdv - seq_dvdh + seq_dtdvdh
    seq_vh = seq_dt - seq_dtdv - seq_dtdh + seq_dtdvdh
    seq_tvh = seq - (seq_t + seq_v + seq_h + seq_tv + seq_th + seq_vh)
    seqs = seq_t, seq_v, seq_h, seq_tv, seq_th, seq_vh, seq_tvh, seq
    return seqs + (NAMES+('tot', ), ) if names else seqs


def get_noise_seqs(seq, method='fast', names=False):
    """
    Wrapper function to get the 3D noise sequences
    """
    if method == "fast":
        return get_all_3d_noise_seq_fast(seq, names=names)
    elif method =="classic":
        return get_all_3d_noise_seq_classic(seq, names=names)
    else:
        raise ValueError("method must be 'fast' or 'classic'")


## matrix approach
def n_dt(seq):
    """
    Applies a temporal mean to the 3D input sequence.
    This operator is used in a matrix-approach to 
    compute all 7 noise sequences.
    """
    return dt(seq)


def n_dv(seq):
    """
    Applies a vertical mean to the 3D input sequence.
    This operator is used in a matrix-approach to 
    compute all 7 noise sequences.
    """
    return dv(seq)


def n_dh(seq):
    """
    Applies a horizontal mean to the 3D input sequence.
    This operator is used in a matrix-approach to 
    compute all 7 noise sequences.
    """
    return dh(seq)


def n_dtdv(seq):
    """
    Applies a temporal-vertical mean to the 3D input sequence.
    This operator is used in a matrix-approach to 
    compute all 7 noise sequences.
    """
    return dt(dv(seq))


def n_dtdh(seq):
    """
    Applies a temporal-horizontal mean to the 3D input sequence.
    This operator is used in a matrix-approach to 
    compute all 7 noise sequences.
    """
    return dt(dh(seq))


def n_dvdh(seq):
    """
    Applies a vertical-horizontal mean to the 3D input sequence.
    This operator is used in a matrix-approach to 
    compute all 7 noise sequences.
    """
    return dv(dh(seq))


def get_all_3D_mean_seq(seq, names=False):
    """
    Returns 6 filtered sequences in the following order :
        dvdh, dtdh, dtdv, dh, dv, dt
    This operator is used in a matrix-approach to 
    compute all 7 noise sequences.
    """
    seqs = n_dvdh(seq), n_dtdh(seq), n_dtdv(seq), n_dh(seq), n_dv(seq), n_dt(seq)
    return seqs + (MEAN_NAMES, ) if names else seqs

    
# All these are for debugging/math verifications
#def var_UBO_t(seq):
#    T, V, H = seq.shape
#    return 1/(T-1)* np.sum(dv(dh(idt(seq)))**2)
#def var_UBO_v(seq):
#    T, V, H = seq.shape
#    return 1/(V-1)* np.sum(dt(dh(idv(seq)))**2)
#def var_UBO_h(seq):
#    T, V, H = seq.shape
#    return 1/(H-1)* np.sum(dt(dv(idh(seq)))**2)
#def var_UBO_tv(seq):
#    T, V, H = seq.shape
#    return 1/(T-1)*1/(V-1) * np.sum(dh(idt(idv(seq)))**2)
#def var_UBO_th(seq):
#    T, V, H = seq.shape
#    return 1/(T-1)*1/(H-1) * np.sum(dv(idt(idh(seq)))**2)
#def var_UBO_vh(seq):
#    T, V, H = seq.shape
#    return 1/(H-1)*1/(V-1) * np.sum(dt(idv(idh(seq)))**2)
#def var_UBO_tvh(seq):
#    T, V, H = seq.shape
#    return 1/(T-1)*1/(V-1)*1/(H-1) * np.sum(idt(idv(idh(seq)))**2)
#
#
#def get_3D_UBO_var(seq):
#    return var_UBO_t(seq), var_UBO_v(seq), var_UBO_h(seq), var_UBO_tv(seq), var_UBO_th(seq), var_UBO_vh(seq), var_UBO_tvh(seq)
#
#
#def get_3D_UBO_corrected_var(seq):
#    T, V, H = seq.shape
#    
#    var_UBO_tvh_val = var_UBO_tvh(seq)
#    
#    var_UBO_vh_val = var_UBO_vh(seq) - 1/T * var_UBO_tvh_val
#    var_UBO_th_val = var_UBO_th(seq) - 1/T * var_UBO_tvh_val
#    var_UBO_tv_val = var_UBO_tv(seq) - 1/T * var_UBO_tvh_val
#    
#    var_UBO_t_val = var_UBO_t(seq) - 1/(V*H)*var_UBO_tvh_val - 1/V * var_UBO_th_val - 1/H * var_UBO_tv_val
#    var_UBO_v_val = var_UBO_v(seq) - 1/(V*T)*var_UBO_tvh_val - 1/V * var_UBO_vh_val - 1/T * var_UBO_tv_val
#    var_UBO_h_val = var_UBO_h(seq) - 1/(H*T)*var_UBO_tvh_val - 1/T * var_UBO_th_val - 1/H * var_UBO_vh_val
#    
#    vec_var = var_UBO_t_val, var_UBO_v_val, var_UBO_h_val, var_UBO_tv_val, var_UBO_th_val, var_UBO_vh_val, var_UBO_tvh_val
#    
#    var_tot = np.sum(vec_var)
#    return vec_var, var_tot


import numpy as np
import unittest

from noise3d import opr

class TestGenseq(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        cls.T = 100
        cls.V = 100
        cls.H = 100
        vals = np.random.normal(0, 1, cls.T*cls.V*cls.H)
        cls.seq = np.reshape(vals, (cls.T,cls.V,cls.H))
            
    def test_10_base_operators(cls):
        # checking dims and shapes
        cls.assertEqual(opr.dt(cls.seq).shape, (1, cls.V, cls.H))
        cls.assertEqual(opr.dv(cls.seq).shape, (cls.T, 1, cls.H))
        cls.assertEqual(opr.dh(cls.seq).shape, (cls.T, cls.V, 1))
        
    def test_20_base_i_operators(cls):
        # checking dims and shapes
        cls.assertEqual(opr.idt(cls.seq).shape, (cls.T, cls.V, cls.H))
        cls.assertEqual(opr.idv(cls.seq).shape, (cls.T, cls.V, cls.H))
        cls.assertEqual(opr.idh(cls.seq).shape, (cls.T, cls.V, cls.H))
        
        
    def test_30_noise_seq_extraction(cls):
        # checking dims and shapes
        # 1d noise seqs
        cls.assertEqual(opr.n_t(cls.seq).shape, (cls.T, 1, 1))
        cls.assertEqual(opr.n_v(cls.seq).shape, (1, cls.V, 1))
        cls.assertEqual(opr.n_h(cls.seq).shape, (1, 1, cls.H))
        # 2d noise seqs
        cls.assertEqual(opr.n_tv(cls.seq).shape, (cls.T, cls.V, 1))
        cls.assertEqual(opr.n_th(cls.seq).shape, (cls.T, 1, cls.H))
        cls.assertEqual(opr.n_vh(cls.seq).shape, (1, cls.V, cls.H))
        # 3d noise seq
        cls.assertEqual(opr.n_tvh(cls.seq).shape, (cls.T, cls.V, cls.H))
        
    def test_40_noise_seq_basic_extraction_function(cls):
        seqs_basic = opr.get_all_3d_noise_seq_classic(cls.seq)

        for seq_basic in seqs_basic:
            # checking dims and shapes
            cls.assertTrue(seq_basic.shape, (cls.T, cls.V, cls.H))
        
        # checking variances
        cls.assertTrue(np.var(seqs_basic[-1]),
                       np.var(sum(seqs_basic[:-1])))
        
        # False because noises cancels themselves ?
        #cls.assertTrue(np.allclose(sum(seqs_basic[:-1]), seqs_basic[-1]))
        
    def test_45_noise_seq_fast_extraction_function(cls):
        seqs_fast = opr.get_all_3d_noise_seq_fast(cls.seq)

        for seq_fast in seqs_fast:
            # checking dims and shapes
            cls.assertTrue(seq_fast.shape, (cls.T, cls.V, cls.H))
        
        # checking variances
        cls.assertTrue(np.var(seq_fast[-1]),
                       np.var(sum(seq_fast[:-1])))
        
        # False because noises cancels themselves ?
        #cls.assertTrue(np.allclose(sum(seqs_basic[:-1]), seqs_basic[-1]))
        
    def test_50_equivalence_basic_fast_methods(cls):
        seqs_basic = opr.get_all_3d_noise_seq_classic(cls.seq)
        seqs_fast = opr.get_all_3d_noise_seq_fast(cls.seq)
        for seq_basic, seq_fast in zip(seqs_basic, seqs_fast):
            # checking dims and shapes
            cls.assertTrue(seq_basic.shape, seq_fast.shape)
            # checking variances
            cls.assertTrue(np.var(seq_basic), np.var(seq_fast))
            # checking almost equal : False for one seq, because of rounding ?
            # cls.assertTrue(np.allclose(seq_basic, seq_fast))
        
if __name__ == "__main__":
    unittest.main()
        
        


import numpy as np

from .opr import dt, dv, dh, idt, idv, idh
from .opr import n_s, n_t, n_v, n_h, n_tv, n_th, n_vh, n_tvh
from .opr import n_dt, n_dv, n_dh, n_dtdv, n_dtdh, n_dvdh
from .opr import NAMES


DTYPE = np.float64
DDOF = 1


## classic approach
def var_s(seq, ddof=DDOF):
    """Return the variance of the mean sequence."""
    return np.var(n_s(seq), dtype=DTYPE,  ddof=ddof)


def var_nt(seq, ddof=DDOF):
    """Return the t noise sequence variance."""
    return np.var(n_t(seq), dtype=DTYPE,  ddof=ddof)


def var_nv(seq, ddof=DDOF):
    """Return the v noise sequence variance."""
    return np.var(n_v(seq), dtype=DTYPE,  ddof=ddof)


def var_nh(seq, ddof=DDOF):
    """Return the h noise sequence variance."""
    return np.var(n_h(seq), dtype=DTYPE,  ddof=ddof)


def var_ntv(seq, ddof=DDOF):
    """Return the tv noise sequence variance."""
    return np.var(n_tv(seq), dtype=DTYPE,  ddof=ddof)


def var_nth(seq, ddof=DDOF):
    """Return the th noise sequence variance."""
    return np.var(n_th(seq), dtype=DTYPE,  ddof=ddof)


def var_nvh(seq, ddof=DDOF):
    """Return the vh noise sequence variance."""
    return np.var(n_vh(seq), dtype=DTYPE,  ddof=ddof)


def var_ntvh(seq, ddof=DDOF):
    """Return the tvh noise sequence variance."""
    return np.var(n_tvh(seq), dtype=DTYPE,  ddof=ddof)


def get_all_3d_noise_var_classic(seq, names=False):
    """Return all the 7 noise variance and total variance.
    
    Equivalent to extract all 7 noise sequences and take 
    their variances.
    Returns 8 noise variances (7 3D noise component and 
    total variance) in a tuple.
    If names==True, also returns a tuple of strings 
    containing the variances names.
    """
    res = var_nt(seq), var_nv(seq), var_nh(seq), var_ntv(seq), var_nth(seq), var_nvh(seq), var_ntvh(seq), np.var(seq, dtype=DTYPE, ddof=DDOF)
    return res + (NAMES+('tot', ), ) if names else res


## matrix approach
def var_dt(seq, ddof=DDOF):
    """Return dt noise variance.
    
    Used for the matrix approach."""
    return np.var(n_dt(seq), dtype=DTYPE,  ddof=ddof)


def var_dv(seq, ddof=DDOF):
    """Return dv noise variance.
    
    Used for the matrix approach."""
    return np.var(n_dv(seq), dtype=DTYPE,  ddof=ddof)


def var_dh(seq, ddof=DDOF):
    """Return dh noise variance.
    
    Used for the matrix approach."""
    return np.var(n_dh(seq), dtype=DTYPE,  ddof=ddof)


def var_dtdv(seq, ddof=DDOF):
    """Return dtdv noise variance.
    
    Used for the matrix approach."""
    return np.var(n_dtdv(seq), dtype=DTYPE,  ddof=ddof)


def var_dtdh(seq, ddof=DDOF):
    """Return dtdh noise variance.
    
    Used for the matrix approach."""
    return np.var(n_dtdh(seq), dtype=DTYPE,  ddof=ddof)


def var_dvdh(seq, ddof=DDOF):
    """Return dvdh noise variance.
    
    Used for the matrix approach."""
    return np.var(n_dvdh(seq), dtype=DTYPE,  ddof=ddof)


def var_tot(seq, ddof=DDOF):
    """Return the sequence variance"""
    return np.var(seq, dtype=DTYPE,  ddof=ddof)


def get_all_3d_mean_var(seq, names=False):
    """Return all 7 mean noise variance of the matrix approach.
    
    The 6 mean and the total variance."""
    res = var_dvdh(seq), var_dtdh(seq), var_dtdv(seq), var_dh(seq), var_dv(seq), var_dt(seq), var_tot(seq)
    return res + (MEAN_NAMES + ('tot', ), ) if names else res


## Fast compute
def get_all_3d_noise_var_fast(seq, ddof=DDOF, names=False):
    """Faster and equivalent way of getting the 3D noise variances."""
    seq_dt = dt(seq)
    seq_dv = dv(seq)
    seq_dh = dh(seq)
    seq_dtdv = dv(seq_dt)
    seq_dtdh = dh(seq_dt)
    seq_dvdh = dv(seq_dh)
    
    var_nt = np.var(seq_dvdh, dtype=DTYPE, ddof=ddof)
    var_nv = np.var(seq_dtdh, dtype=DTYPE, ddof=ddof)
    var_nh = np.var(seq_dtdv, dtype=DTYPE, ddof=ddof)
    
    var_nth = np.var(seq_dv - seq_dtdv - seq_dvdh, dtype=DTYPE, ddof=ddof)
    var_ntv = np.var(seq_dh - seq_dvdh - seq_dtdh, dtype=DTYPE, ddof=ddof)
    var_nvh = np.var(seq_dt - seq_dtdv - seq_dtdh, dtype=DTYPE, ddof=ddof)
    
    tot = np.var(seq, dtype=DTYPE, ddof=ddof)
    var_ntvh = tot - (var_nt + var_nv + var_nh + var_ntv + var_nth + var_nvh)
    res = var_nt, var_nv, var_nh, var_ntv, var_nth, var_nvh, var_ntvh, tot
    return res + (NAMES + ("tot", ), ) if names else res
    

# MATRIX APPROACH
# classic mixin matrix
M_classic = np.array([
[ 1,  0,  0,  0,  0,  0,  0],
[ 0,  1,  0,  0,  0,  0,  0],
[ 0,  0,  1,  0,  0,  0,  0],
[ 1,  1,  0,  1,  0,  0,  0],
[ 1,  0,  1,  0,  1,  0,  0],
[ 0,  1,  1,  0,  0,  1,  0],
[ 1,  1,  1,  1,  1,  1,  1],   
])



def compute_M_corrected(T, V, H, VH=None):
    """
    Compute unbiased mixin matrix
    """
    if VH is None:
        VH = V*H
    # if not mask is used, VH should be equal to V*H
    mat = np.array([
    [                1,                   0,                  0,                 1/V,                 1/H,                   0,  1/(VH)],
    [                0,                   1,                  0,                 1/T,                   0,                 1/H, 1/(T*H)],
    [                0,                   0,                  1,                   0,                 1/T,                 1/V, 1/(T*V)],
    [  V*(T-1)/(T*V-1),     T*(V-1)/(T*V-1),                  0,                   1, V*(T-1)/((T*V-1)*H), T*(V-1)/((T*V-1)*H),     1/H],
    [  H*(T-1)/(T*H-1),                   0,    T*(H-1)/(T*H-1), H*(T-1)/((T*H-1)*V),                   1, T*(H-1)/((T*H-1)*V),     1/V],
    [                0,      H*(V-1)/(VH-1),     V*(H-1)/(VH-1),  H*(V-1)/((VH-1)*T),  V*(H-1)/((VH-1)*T),                   1,     1/T],
    [VH*(T-1)/(T*VH-1),  T*H*(V-1)/(T*VH-1), T*V*(H-1)/(T*VH-1),  H*(T*V-1)/(T*VH-1),  V*(T*H-1)/(T*VH-1),   T*(VH-1)/(T*VH-1),       1],
    ], dtype=float)
    return mat


def _get_all_3d_variance_from_matrix(seq, M):
    """
    Generic function to compute noise variances from mean variances.
    This is a helper function for :
     - get_all_3d_classic_var_matrix
     - get_all_3d_corrected_var_matrix
    """
    vec_var_D = np.array(get_all_3d_mean_var(seq))
    M_inv = np.linalg.inv(M)
    vec_var_sigma = tuple(np.matmul(M_inv, vec_var_D))
    return vec_var_sigma + (sum(vec_var_sigma),)
    

def get_all_3d_classic_var_matrix(seq, names=False):
    """
    Compute classic noise variances using the matrix mixin approach.
    """
    res =  _get_all_3d_variance_from_matrix(seq, M_classic)
    return res + (NAMES + ('tot', ), ) if names else res


def get_all_3d_corrected_var_matrix(seq, names=False):
    """
    Compute unbiased noise variances using the matrix mixin approach.
    """
    Tv, Vv, Hv, VHv = get_valid_counts(seq)
    M_corrected = compute_M_corrected(Tv, Vv, Hv, VHv)
    res = _get_all_3d_variance_from_matrix(seq, M_corrected)
    return res + (NAMES + ('tot', ), ) if names else res


def get_noise_vars(seq, method="fast", ddof=DDOF, names=False):
    if method == "fast":
        return get_all_3d_noise_var_fast(seq, ddof=DDOF, names=names)
    elif method == "classic":
        return get_all_3d_noise_var_classic(seq, ddof=DDOF, names=names)
    elif method == "matrix_classic":
        return get_all_3d_classic_var_matrix(seq, names=names)
    elif method == "matrix_corrected":
        return get_all_3d_corrected_var_matrix(seq, names=names)
    else:
        raise NotImplementedError("Method must be among classic, fast, matrix_classic and matrix_corrected")


def get_valid_counts(seq):
    # shapes
    Ts, Vs, Hs = seq.shape 
    # valid counts
    #Tv, Vv, Hv = (np.ma.count(seq, axis=0), np.ma.count(seq, axis=1), np.ma.count(seq, axis=2))
    #print(Tv.shape)
    #print(Vv.shape)
    #print(Hv.shape)
    # valid counts in 2D plane
    # if not mask is used, VH should be equal to V*H
    VHv = np.ma.count(seq[0]) # count the valid values on first image
    return Ts, Vs, Hs, VHv
    

# Classical noise estimators
def var_netd(seq, axis=0, ddof=1):
    """
    Compute the NETD variance the "usual" way.
    """
    return np.mean(np.var(seq, axis=axis, ddof=ddof))


def var_fpn(seq, axis=0, ddof=1):
    """
    Compute the fpn variance the 'usual' way.
    """
    return np.var(np.mean(seq, axis=axis), ddof=ddof)


# For debugging/testing
#def get_all_3d_UBO_var_matrix(seq):
#    """
#    Compute UBO noise variances using the matrix mixin approach.
#    """
#    Tv, Vv, Hv, VHv = get_valid_counts(seq)    
#    M_UBO = compute_M_UBO(Tv, Vv, Hv, VHv)
#    return _get_all_3d_variance_from_matrix(seq, M_UBO)
#
#
#def compute_M_UBO(T, V, H, VH=None):
#    """
#    Compute mixin matrix from paper ####
#    """
#    if VH is None:
#        VH = V*H
#    # if not mask is used, VH should be equal to V*H
#    mat = np.array([
#    [ 1,  0,  0,             1/V,             1/H,              0,                                  1/(VH)],
#    [ 0,  1,  0,             1/T,               0,             1/H,                                 1/(T*H)],
#    [ 0,  0,  1,               0,             1/T,             1/V,                                 1/(T*V)],
#    [ 1,  1,  0, (V+T+T*V)/(T*V),             1/H,             1/H,                   (T + V + T*V)/(T*VH)],
#    [ 1,  0,  1,             1/V, (H+T+T*H)/(T*H),             1/V,                       (H+T+H*T)/(T*VH)],
#    [ 0,  1,  1,             1/T,             1/T, (H+V+H*V)/(VH),                       (H+V+H*V)/(T*VH)],
#    [ 1,  1,  1, (V+T+T*V)/(T*V), (H+T+T*H)/(T*H), (H+V+H*V)/(VH), (H+T+V+ T*V + H*T + VH + T*VH)/(T*VH)]
#    ])
#
#    return mat



# -*- coding: utf-8 -*-
import os
import pathlib
import re

from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

here = os.path.abspath(os.path.dirname(__file__))

VERSIONFILE="noise3d/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

README = (HERE / "README.md").read_text()

setup(
    name="noise3d",
    version=verstr,
    description="Noise analysis based on 3D noise model",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/mocquin/noise3d",
    author="mocquin",
    author_email="mocquin@me.com",
    license="MIT",
    keywords='noise 3Dnoise',
    packages=find_packages(exclude=("tests", )),
    # add content of MANIFEST
    include_package_data=True,
    install_requires=["numpy", 
                      "scipy", 
                      "matplotlib",
                     ]
)
